# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from abc import ABCMeta, abstractmethod


class UpdaterConsumer(metaclass=ABCMeta):
    """Event consumer

    """
    def __init__(self, backend, batch_cache_write=1000):
        super().__init__()
        self._reset_cache()
        self.backend = backend
        self.batch = int(batch_cache_write)
        logging.basicConfig(level=logging.DEBUG)
        self.log = logging.getLogger('%s.%s' % (
            self.__class__.__module__, self.__class__.__name__))

    def _reset_cache(self):
        """Reset internal cache.

        """
        self.count = 0
        self.seen_events = set()
        self.events = []

    def is_interesting(self, event):
        """Determine if an event is interesting or not.

        Args:
            event (SWHEvent): SWH event

        """
        return event.is_interesting()

    @abstractmethod
    def convert_event(self, event):
        """Parse an event into an SWHEvent.

        """
        pass

    def process_event(self, event):
        """Process converted and interesting event.

        Args:
            event (SWHEvent): Event to process if deemed interesting

        """
        try:
            if event.url in self.seen_events:
                event.cnt += 1
            else:
                self.events.append(event)
                self.seen_events.add(event.url)
                self.count += 1
        finally:
            if self.count >= self.batch:
                if self.events:
                    self.backend.cache_put(self.events)
                self._reset_cache()

    def _flush(self):
        """Flush remaining internal cache if any.

        """
        if self.events:
            self.backend.cache_put(self.events)
            self._reset_cache()

    @abstractmethod
    def has_events(self):
        """Determine if there remains events to consume.

        Returns
            boolean value, true for remaining events, False otherwise

        """
        pass

    @abstractmethod
    def consume_events(self):
        """The main entry point to consume events.

        This should either yield or return message for consumption.

        """
        pass

    @abstractmethod
    def open_connection(self):
        """Open a connection to the remote system we are supposed to consume
           from.

        """
        pass

    @abstractmethod
    def close_connection(self):
        """Close opened connection to the remote system.

        """
        pass

    def run(self):
        """The main entry point to consume events.

        """
        try:
            self.open_connection()
            while self.has_events():
                for _event in self.consume_events():
                    event = self.convert_event(_event)
                    if not event:
                        self.log.warning(
                            'Incomplete event dropped %s' % _event)
                        continue
                    if not self.is_interesting(event):
                        continue
                    if self.debug:
                        self.log.debug('Event: %s' % event)
                    try:
                        self.process_event(event)
                    except Exception:
                        self.log.exception(
                            'Problem when processing event %s' % _event)
                        continue
        except Exception as e:
            self.log.error('Error raised during consumption: %s' % e)
            raise e
        finally:
            self.close_connection()
            self._flush()



import os
import pytest
import glob
from datetime import timedelta

from swh.core.utils import numfile_sortkey as sortkey
from swh.scheduler import get_scheduler
from swh.scheduler.tests import SQL_DIR

# make sure we are not fooled by CELERY_ config environment vars
for var in [x for x in os.environ.keys() if x.startswith('CELERY')]:
    os.environ.pop(var)

import swh.scheduler.celery_backend.config  # noqa
# this import is needed here to enforce creation of the celery current app
# BEFORE the swh_app fixture is called, otherwise the Celery app instance from
# celery_backend.config becomes the celery.current_app


# test_cli tests depends on a en/C locale, so ensure it
os.environ['LC_ALL'] = 'C.UTF-8'

DUMP_FILES = os.path.join(SQL_DIR, '*.sql')

# celery tasks for testing purpose; tasks themselves should be
# in swh/scheduler/tests/tasks.py
TASK_NAMES = ['ping', 'multiping', 'add', 'error']


@pytest.fixture(scope='session')
def celery_enable_logging():
    return True


@pytest.fixture(scope='session')
def celery_includes():
    return [
        'swh.scheduler.tests.tasks',
    ]


@pytest.fixture(scope='session')
def celery_parameters():
    return {
        'task_cls': 'swh.scheduler.task:SWHTask',
        }


@pytest.fixture(scope='session')
def celery_config():
    return {
        'accept_content': ['application/x-msgpack', 'application/json'],
        'task_serializer': 'msgpack',
        'result_serializer': 'json',
        }


# override the celery_session_app fixture to monkeypatch the 'main'
# swh.scheduler.celery_backend.config.app Celery application
# with the test application.
@pytest.fixture(scope='session')
def swh_app(celery_session_app):
    swh.scheduler.celery_backend.config.app = celery_session_app
    yield celery_session_app


@pytest.fixture
def swh_scheduler(request, postgresql_proc, postgresql):
    scheduler_config = {
        'db': 'postgresql://{user}@{host}:{port}/{dbname}'.format(
            host=postgresql_proc.host,
            port=postgresql_proc.port,
            user='postgres',
            dbname='tests')
    }

    all_dump_files = sorted(glob.glob(DUMP_FILES), key=sortkey)

    cursor = postgresql.cursor()
    for fname in all_dump_files:
        with open(fname) as fobj:
            cursor.execute(fobj.read())
    postgresql.commit()

    scheduler = get_scheduler('local', scheduler_config)
    for taskname in TASK_NAMES:
        scheduler.create_task_type({
            'type': 'swh-test-{}'.format(taskname),
            'description': 'The {} testing task'.format(taskname),
            'backend_name': 'swh.scheduler.tests.tasks.{}'.format(taskname),
            'default_interval': timedelta(days=1),
            'min_interval': timedelta(hours=6),
            'max_interval': timedelta(days=12),
        })

    return scheduler



# Copyright (C) 2015-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import current_app
import celery.app.task
from celery.utils.log import get_task_logger

from swh.core.statsd import Statsd


class SWHTask(celery.app.task.Task):
    """a schedulable task (abstract class)

    Current implementation is based on Celery. See
    http://docs.celeryproject.org/en/latest/reference/celery.app.task.html for
    how to use tasks once instantiated

    """

    _statsd = None
    _log = None

    @property
    def statsd(self):
        if self._statsd:
            return self._statsd
        worker_name = current_app.conf.get('worker_name')
        if worker_name:
            self._statsd = Statsd(constant_tags={
                'task': self.name,
                'worker': worker_name,
            })
            return self._statsd
        else:
            return Statsd(constant_tags={
                'task': self.name,
                'worker': 'unknown worker',
            })

    def __call__(self, *args, **kwargs):
        self.statsd.increment('swh_task_called_count')
        with self.statsd.timed('swh_task_duration_seconds'):
            return super().__call__(*args, **kwargs)

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        self.statsd.increment('swh_task_failure_count')

    def on_success(self, retval, task_id, args, kwargs):
        self.statsd.increment('swh_task_success_count')
        # this is a swh specific event. Used to attach the retval to the
        # task_run
        self.send_event('task-result', result=retval)

    @property
    def log(self):
        if self._log is None:
            self._log = get_task_logger(self.name)
        return self._log

    def run(self, *args, **kwargs):
        self.log.debug('%s: args=%s, kwargs=%s', self.name, args, kwargs)
        ret = super().run(*args, **kwargs)
        self.log.debug('%s: OK => %s', self.name, ret)
        return ret



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import patch

from hypothesis import given
from hypothesis.strategies import sampled_from

from swh.scheduler.updater.events import SWHEvent
from swh.scheduler.updater.ghtorrent import (INTERESTING_EVENT_KEYS,
                                             GHTorrentConsumer, events)
from swh.scheduler.updater.backend import SchedulerUpdaterBackend

from . import UpdaterTestUtil, from_regex


def event_values():
    return set(events['evt']).union(set(events['ent']))


def ghtorrentize_event_name(event_name):
    return '%sEvent' % event_name.capitalize()


EVENT_TYPES = sorted([ghtorrentize_event_name(e) for e in event_values()])


class FakeChannel:
    """Fake Channel (virtual connection inside a connection)

    """
    def close(self):
        self.close = True


class FakeConnection:
    """Fake Rabbitmq connection for test purposes

    """
    def __init__(self, conn_string):
        self._conn_string = conn_string
        self._connect = False
        self._release = False
        self._channel = False

    def connect(self):
        self._connect = True
        return True

    def release(self):
        self._connect = False
        self._release = True

    def channel(self):
        self._channel = True
        return FakeChannel()


class GHTorrentConsumerTest(UpdaterTestUtil, unittest.TestCase):
    def setUp(self):
        config = {
            'ghtorrent': {
                'rabbitmq': {
                    'conn': {
                        'url': 'amqp://u:p@https://somewhere:9807',
                    },
                    'prefetch_read': 17,
                },
                'batch_cache_write': 42,
            },
            'scheduler_updater': {
                'cls': 'local',
                'args': {
                    'db': 'dbname=softwareheritage-scheduler-updater-dev',
                },
            },
        }

        GHTorrentConsumer.connection_class = FakeConnection
        with patch.object(
                SchedulerUpdaterBackend, '__init__', return_value=None):
            self.consumer = GHTorrentConsumer(**config)

    @patch('swh.scheduler.updater.backend.SchedulerUpdaterBackend')
    def test_init(self, mock_backend):
        # given
        # check init is ok
        self.assertEqual(self.consumer.batch, 42)
        self.assertEqual(self.consumer.prefetch_read, 17)

    def test_has_events(self):
        self.assertTrue(self.consumer.has_events())

    def test_connection(self):
        # when
        self.consumer.open_connection()

        # then
        self.assertEqual(self.consumer.conn._conn_string,
                         'amqp://u:p@https://somewhere:9807')
        self.assertTrue(self.consumer.conn._connect)
        self.assertFalse(self.consumer.conn._release)

        # when
        self.consumer.close_connection()

        # then
        self.assertFalse(self.consumer.conn._connect)
        self.assertTrue(self.consumer.conn._release)
        self.assertIsInstance(self.consumer.channel, FakeChannel)

    @given(sampled_from(EVENT_TYPES),
           from_regex(r'^[a-z0-9]{5,7}/[a-z0-9]{3,10}$'))
    def test_convert_event_ok(self, event_type, name):
        input_event = self._make_event(event_type, name, 'git')
        actual_event = self.consumer.convert_event(input_event)

        self.assertTrue(isinstance(actual_event, SWHEvent))

        event = actual_event.get()

        expected_event = {
            'type': event_type.lower().rstrip('Event'),
            'url': 'https://github.com/%s' % name,
            'last_seen': input_event['created_at'],
            'cnt': 1,
            'origin_type': 'git',
        }
        self.assertEqual(event, expected_event)

    @given(sampled_from(EVENT_TYPES),
           from_regex(r'^[a-z0-9]{5,7}/[a-z0-9]{3,10}$'),
           sampled_from(INTERESTING_EVENT_KEYS))
    def test_convert_event_ko(self, event_type, name, missing_data_key):
        input_event = self._make_incomplete_event(
            event_type, name, 'git', missing_data_key)

        logger = self.consumer.log
        del self.consumer.log  # prevent gazillions of warnings
        actual_converted_event = self.consumer.convert_event(input_event)
        self.consumer.log = logger
        self.assertIsNone(actual_converted_event)

    @patch('swh.scheduler.updater.ghtorrent.collect_replies')
    def test_consume_events(self, mock_collect_replies):
        # given
        self.consumer.queue = 'fake-queue'  # hack
        self.consumer.open_connection()

        fake_events = [
            self._make_event('PushEvent', 'user/some-repo', 'git'),
            self._make_event('PushEvent', 'user2/some-other-repo', 'git'),
        ]

        mock_collect_replies.return_value = fake_events

        # when
        actual_events = []
        for e in self.consumer.consume_events():
            actual_events.append(e)

        # then
        self.assertEqual(fake_events, actual_events)

        mock_collect_replies.assert_called_once_with(
            self.consumer.conn,
            self.consumer.channel,
            'fake-queue',
            no_ack=False,
            limit=17
        )



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click


@click.command()
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
@click.pass_context
def main(ctx, verbose):
    """Consume events from ghtorrent and write them to cache.

    """
    click.echo("Deprecated! Use 'swh-scheduler updater' instead.",
               err=True)
    ctx.exit(1)


if __name__ == '__main__':
    main()
